package com.bergerkiller.bukkit.nolagg.meta;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.LongHashMap;
import net.minecraft.server.PlayerManager;
import net.minecraft.server.WorldServer;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.craftbukkit.entity.CraftEntity;
import org.bukkit.entity.Entity;

import com.bergerkiller.bukkit.common.SafeField;
import com.bergerkiller.bukkit.common.utils.CommonUtil;
import com.bergerkiller.bukkit.common.utils.EntityUtil;
import com.bergerkiller.bukkit.nolagg.NoLagg;
import com.bergerkiller.bukkit.nolagg.monitor.PerformanceMonitor;
import com.bergerkiller.bukkit.nolagg.spawnlimit.SpawnHandler;

public class WorldMetaData {

	private static final List<WorldMetaData> worlds = new ArrayList<WorldMetaData>();
	private static final HashMap<World, WorldMetaData> worlddata = new HashMap<World, WorldMetaData>();
	public static void deinit() {
		synchronized (worlddata) {
			for (WorldMetaData dat : worlddata.values()) {
				dat.remove();
			}
			worlddata.clear();
			worlds.clear();
		}
	}
	public static void remove(final World world) {
		synchronized (worlddata) {
			WorldMetaData dat = worlddata.remove(world);
			if (dat != null) {
				worlds.remove(dat);
				dat.remove();
			}
		}
	}
	public static WorldMetaData get(final World world) {
		if (world == null) return null;
		synchronized (worlddata) {
			WorldMetaData dat = worlddata.get(world);
			if (dat == null) {
				dat = new WorldMetaData(world);
				worlddata.put(world, dat);
				worlds.add(dat);
			}
			return dat;
		}
	}
	public static WorldMetaData get(final net.minecraft.server.World world) {
		return get(world.getWorld());
	}
	public static List<WorldMetaData> getWorlds() {
		return worlds;
	}
	public static void updateAll() {
		SpawnHandler.reset();
		synchronized (worlddata) {
			if (NoLagg.isMonitorEnabled) {
				PerformanceMonitor.entityListGen.start();
			}
			for (WorldMetaData dat : worlds) dat.updateEntities();	
			if (NoLagg.isMonitorEnabled) {
				PerformanceMonitor.entityListGen.next();
			}

			if (NoLagg.useSpawnLimits) {
				if (NoLagg.isMonitorEnabled) {
					PerformanceMonitor.entitySpawnLimit.start();	
				}
				for (WorldMetaData dat : worlds) dat.updateSpawnLimits();
		     	if (NoLagg.isMonitorEnabled) {
					PerformanceMonitor.entitySpawnLimit.next();
				}
			}
			
			if (NoLagg.isMonitorEnabled) {
				PerformanceMonitor.chunkUpdate.start();
			}
            for (WorldMetaData dat : worlds) dat.updateChunks();
            if (NoLagg.isMonitorEnabled) {						
                PerformanceMonitor.chunkUpdate.next();
            }
		}
	}
	private void updateEntities() {
		//update entity list
		this.entitylist.clear();
		CraftEntity e;
		for (Object obj : this.getNative().entityList) {
			e = (CraftEntity) ((net.minecraft.server.Entity) obj).getBukkitEntity();
			if (e.isDead()) continue;
			this.entitylist.add(e);
		}
		Collections.reverse(this.entitylist); //make sure the oldest entity is at front
	}
	private void updateSpawnLimits() {
		SpawnHandler.update(this.world, this.entitylist);
	}

	public static int getSpawnChunkX(WorldServer world) {
		return world.worldData.c() >> 4;
	}
	public static int getSpawnChunkZ(WorldServer world) {
		return world.worldData.e() >> 4;
	}
	public int getSpawnChunkX() {
		return getSpawnChunkX(this.getNative());
	}
	public int getSpawnChunkZ() {
		return getSpawnChunkZ(this.getNative());
	}

	private void updateChunks() {
		//update chunks and unload them if needed
		synchronized (this.chunkdata) {
			if (!this.chunkdata.isEmpty()) {
				final long time = System.currentTimeMillis() - ChunkMetaData.unloadDelay;
				Iterator<ChunkMetaData> it = this.chunkdata.values().iterator();
				ChunkMetaData dat;
				WorldServer world = this.getNative();
				while (it.hasNext()) {
					dat = it.next();
					if (!dat.isLoaded()) {
						it.remove();
					} else if (world.savingDisabled || !dat.isUnloadable() || dat.isPlayerNear()) {
						dat.touch();
					} else if (dat.isExpired(time)) {
						this.getChunkProvider().queueUnload(dat.getX(), dat.getZ());
					} else {
						dat.markNotUnloadable();
					}
				}

			}
		}
	}

	public static int getTotalChunkCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) c += wmd.getChunkCount();	
		}
		return c;
	}
	public static int getTotalUnloadableChunkCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) {
				synchronized (wmd.chunkdata) {
					for (ChunkMetaData cmd : wmd.chunkdata.values()) {
						if (cmd.isUnloadable()) c++;
					}
				}
			}
		}
		return c;
	}

	private static SafeField<LongHashMap> managersField;
	private WorldMetaData(final World world) {
		this.world = (CraftWorld) world;
		if (managersField == null) {
			managersField = new SafeField<LongHashMap>(PlayerManager.class, "b");
		}
		this.players = managersField.get(this.getNative().manager);
		this.chunkdata = new LinkedHashMap<Long, ChunkMetaData>(CommonUtil.chunkArea);
		this.entitylist = new ArrayList<Entity>();
		this.add();
	}

	public final CraftWorld world;
	private final Map<Long, ChunkMetaData> chunkdata;
	private final List<Entity> entitylist;
	private final LongHashMap players;

	public boolean isPlayerNear(ChunkMetaData chunk) {
		if (this.players == null) {
			final int bx = chunk.getX() << 4;
			final int bz = chunk.getZ() << 4;
			net.minecraft.server.Entity e;
			for (Object o : this.getNative().players) {
				e = (net.minecraft.server.Entity) o;
				if (EntityUtil.isNearBlock(e, bx, bz, CommonUtil.blockView)) return true;
			}
			return false;
		} else {
			return this.players.contains(chunk.coord);
		}
	}
	
 	public ChunkProviderServer getChunkProvider() {
		return this.getNative().chunkProviderServer;
	}
	public WorldServer getNative() {
		return this.world.getHandle();
	}

	public List<Entity> getEntities() {
		return this.entitylist;
	}

	public ChunkMetaData getChunk(final Location location) {
		return this.getChunk(location.getBlockX() >> 4, location.getBlockZ() >> 4);
	}
	public ChunkMetaData getChunk(final net.minecraft.server.Chunk chunk) {
		return this.getChunk(chunk.x, chunk.z);
	}
	public ChunkMetaData getChunk(final Chunk chunk) {
		return this.getChunk(chunk.getX(), chunk.getZ());
	}
	public ChunkMetaData getChunk(final int x, final int z) {
		final long key = ChunkMetaData.getKey(x, z);
		synchronized (this.chunkdata) {
			ChunkMetaData dat = this.chunkdata.get(key);
			if (dat == null && this.getChunkProvider().isChunkLoaded(x, z)) {
				dat = new ChunkMetaData(this, x, z);
				this.chunkdata.put(key, dat);
			}
			return dat;
		}
	}
	public int getChunkCount() {
		synchronized (this.chunkdata) {
			return this.chunkdata.size();
		}
	}

	protected void removeChunk(final ChunkMetaData dat) {
		synchronized (this.chunkdata) {
			this.chunkdata.remove(dat.coord);
		}
	}
	public boolean removeChunk(final int x, final int z) {
		synchronized (this.chunkdata) {
			ChunkMetaData dat = this.chunkdata.get(ChunkMetaData.getKey(x, z));
			if (dat != null) {
				dat.remove();
				return true;
			} else {
				return false;
			}
		}
	}

	private void add() {
		for (Chunk c : world.getLoadedChunks()) getChunk(c);
	}
	private void remove() {
		synchronized (this.chunkdata) {
			Iterator<ChunkMetaData> it = this.chunkdata.values().iterator();
			ChunkMetaData dat;
			while (it.hasNext()) {
				dat = it.next();
				dat.remove(false);
				it.remove();
			}
		}
		//other unload code here
	}

}
